//-------------------------------------------------------------------------------------------------------
// VST Plug-Ins SDK
// Version 2.4		$Date: 2006/11/13 09:08:27 $
//
// Category     : VST 2.x SDK Samples
// Filename     : filters.cpp
// Created by   : Tom Erbe
// Description  : Collection of simple filters
//
// � 2012, Tom Erbe, All Rights Reserved
//-------------------------------------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <cmath>
//#include "midiconstants.h"
#ifndef __filters__
#include "filters.h"
#endif


//-------------------------------------------------------------------------------------------------------
AudioEffect* createEffectInstance (audioMasterCallback audioMaster)
{
	return new Filters (audioMaster);
}


//----------------------------------------------------------------------------- 
FiltersProgram::FiltersProgram ()
{
	// default Program Values
	preParameter[kType] = 0.0f;
	preParameter[kFreq] = 0.5f;
	preParameter[kReson] = 0.5f;

	strcpy (name, "Init");
}

//-----------------------------------------------------------------------------
Filters::Filters (audioMasterCallback audioMaster)
	: AudioEffectX (audioMaster, kNumPrograms, kNumParams)
{
	// init
	long i;
	float peak;
	
	twoPi = 8.0 * atan(1.0);
	Pi = 4.0 * atan(1.0);
	halfPi = 2.0 * atan(1.0);
	programs = new FiltersProgram[numPrograms];
	guiParameter[kFreq] = guiParameter[kType] = guiParameter[kReson] = 0.0f;
	dspParameter[kFreq] = dspParameter[kType] = dspParameter[kReson] = 0.0f;
	
	// build some wavetables - in case i want to add an LFO later
	sineTable = new float[8192];
	for(i = 0; i < 8192; i++)
		*(sineTable + i) = sin(twoPi * (float)i/8192.0f);
	// square wave from odd harmonics
	squareTable = new float[8192];
	for(i = 0; i < 8192; i++)
		*(squareTable + i) = sin(twoPi * (float)i/8192.0f)
		+ 1.0/3.0 * sin(twoPi * (float)(i*3.0)/8192.0f)
		+ 1.0/5.0 * sin(twoPi * (float)(i*5.0)/8192.0f)
		+ 1.0/7.0 * sin(twoPi * (float)(i*7.0)/8192.0f)
		+ 1.0/9.0 * sin(twoPi * (float)(i*9.0)/8192.0f)
		+ 1.0/11.0 * sin(twoPi * (float)(i*11.0)/8192.0f)
		+ 1.0/13.0 * sin(twoPi * (float)(i*13.0)/8192.0f)
		+ 1.0/15.0 * sin(twoPi * (float)(i*15.0)/8192.0f)
		+ 1.0/17.0 * sin(twoPi * (float)(i*17.0)/8192.0f)
		+ 1.0/19.0 * sin(twoPi * (float)(i*19.0)/8192.0f)
		+ 1.0/21.0 * sin(twoPi * (float)(i*21.0)/8192.0f)
		+ 1.0/23.0 * sin(twoPi * (float)(i*23.0)/8192.0f)
		+ 1.0/25.0 * sin(twoPi * (float)(i*25.0)/8192.0f);
	// find the peak and normalize
	peak = 0.0f;
	for(i = 0; i < 8192.0f; i++)
	{
		if(*(squareTable + i) > peak)
			peak = *(squareTable + i);
		if(*(squareTable + i) < -peak)
			peak = -1.0 * *(squareTable + i);
	}
	for(i = 0; i < 8192.0f; i++)
		*(squareTable + i) *= 1.0f/peak;
	// dumb ramp wave with aliasing
	rampTable = new float[8192];
	for(i = 0; i < 8192; i++)
		*(rampTable + i) = ((float)i/(float)8192 * 2.0) - 1.0f;
		
	
	
	if (programs)
		setProgram (0);

	setNumInputs (2);	// mono input
	setNumOutputs (2);	// stereo output

	setUniqueID ('Flrs');	// this should be unique, use the Steinberg web page for plugin Id registration

	resume ();		// flush buffer
	curSample = 0;
}

//------------------------------------------------------------------------
Filters::~Filters ()
{
	if(sineTable)
		delete [] sineTable;
	if(squareTable)
		delete [] squareTable;
	if(rampTable)
		delete [] rampTable;
	if (programs)
		delete[] programs;
}

//------------------------------------------------------------------------
void Filters::setProgram (VstInt32 program)
{
	int i;
	curProgram = program;
	for(i = 0; i < kNumParams; i++)
		setParameter(i, programs[program].preParameter[i]);
}

//------------------------------------------------------------------------
void Filters::setProgramName (char *name)
{
	strcpy (programs[curProgram].name, name);
}

//------------------------------------------------------------------------
void Filters::getProgramName (char *name)
{
	if (!strcmp (programs[curProgram].name, "Init"))
		sprintf (name, "%s %d", programs[curProgram].name, curProgram + 1);
	else
		strcpy (name, programs[curProgram].name);
}

//-----------------------------------------------------------------------------------------
bool Filters::getProgramNameIndexed (VstInt32 category, VstInt32 index, char* text)
{
	if (index < kNumPrograms)
	{
		strcpy (text, programs[index].name);
		return true;
	}
	return false;
}

//------------------------------------------------------------------------
void Filters::resume ()
{
	// reset state variables
	opY1L = opY1R = foX1L = foX1R = foY1L = foY1R = 0.0;
	soX2L = soX1L = soY0L = soY2L = soY1L = 0.0;
	soX2R = soX1R = soY0R = soY2R = soY1R = 0.0;
	bqX2L = bqX1L = bqY2L = bqY1L = bqY0L = 0.0;
	bqX2R = bqX1R = bqY2R = bqY1R = bqY0R = 0.0;
	svLPL = svHPL = svBPL = svBRL = 0.0;
	svLPR = svHPR = svBPR = svBRR = 0.0;
	
	c = d = 0.0;
	
	AudioEffectX::resume ();
}

//------------------------------------------------------------------------
void Filters::setParameter (VstInt32 index, float value)
{
	if(index < kNumParams)
		guiParameter[index] = programs[curProgram].preParameter[index] = value;
}

//------------------------------------------------------------------------
float Filters::getParameter (VstInt32 index)
{
	if(index < kNumParams)
		return(guiParameter[index]);
	else
		return(0.0f);
}

//------------------------------------------------------------------------
void Filters::getParameterName (VstInt32 index, char *label)
{
	switch (index)
	{
		case kType :    
			strcpy (label, "type:");		
			break;
		case kFreq :	
			strcpy (label, "freq:");	
			break;
		case kReson :      
			strcpy (label, "reson:");		
			break;
	}
}

//------------------------------------------------------------------------
void Filters::getParameterDisplay (VstInt32 index, char *text)
{
	float value;
	long lvalue;
	switch (index)
	{
		case kType:
			value = 14.999f * guiParameter[index];	// 15 equal regions on the dial
			lvalue = (long)value;
			switch(lvalue)
			{
				case kType1PoleLP:
					strcpy(text, "1pLP");
					break;
				case kType1PoleHP:
					strcpy(text, "1pHP");
					break;
				case kType1stOrdAP:
					strcpy(text, "foAP");
					break;
				case kType1stOrdLP:
					strcpy(text, "foLP");
					break;
				case kType1stOrdHP:
					strcpy(text, "foHP");
					break;
				case kType2ndOrdAP:
					strcpy(text, "soAP");
					break;
				case kType2ndOrdBR:
					strcpy(text, "soBR");
					break;
				case kType2ndOrdBP:
					strcpy(text, "soBP");
					break;
				case kTypeBQBWLP:
					strcpy(text, "bqLP");
					break;
				case kTypeBQBWHP:
					strcpy(text, "bqHP");
					break;
				case kTypeSVLP:
					strcpy(text, "svLP");
					break;
				case kTypeSVHP:
					strcpy(text, "svHP");
					break;
				case kTypeSVBR:
					strcpy(text, "svBR");
					break;
				case kTypeSVBP:
					strcpy(text, "svBP");
					break;
				case kTypeBypass:
					strcpy(text, "bypass");
					break;
			}
			break;
					
		case kFreq :	
			// 0.0 knob = 20Hz, 1.0 knob = 20,000Hz - log
			// x1000 range
			value = powf(10.0f, (guiParameter[index] * 3.0f) + 1.30103f);
			float2string (value, text, kVstMaxParamStrLen);
			break;
		case kReson : 
			// para 0 to 1
			// -1 to 0
			// -80 to 0
			value = powf(10.0f, (guiParameter[index] * 2.0f - 1.0f));
			float2string (value, text, kVstMaxParamStrLen);
			break;
	}
}

//------------------------------------------------------------------------
void Filters::getParameterLabel (VstInt32 index, char *label)
{
	switch (index)
	{
		case kFreq:
			strcpy (label, "Hz");
			break;
		default :      
			strcpy (label, "");		
			break;
	}
}

//------------------------------------------------------------------------
bool Filters::getEffectName (char* name)
{
	strcpy (name, "filters");
	return true;
}

//------------------------------------------------------------------------
bool Filters::getProductString (char* text)
{
	strcpy (text, "+filters");
	return true;
}

//------------------------------------------------------------------------
bool Filters::getVendorString (char* text)
{
	strcpy (text, "SoundHack");
	return true;
}


//---------------------------------------------------------------------------
void Filters::processReplacing (float** inputs, float** outputs, VstInt32 sampleFrames)
{
	float* inL = inputs[0];
	float* inR = inputs[1];
	float* outL = outputs[0];
	float* outR = outputs[1];
	VstInt32 sample;
	float freqInc, resonInc;
	float linFreq;
	float reson;
	long type;
	
	dspParameter[kType] = guiParameter[kType];
	freqInc = (guiParameter[kFreq] - dspParameter[kFreq])/sampleFrames;
	resonInc = (guiParameter[kReson] - dspParameter[kReson])/sampleFrames;
	type = (long)(14.999f * dspParameter[kType]);	// 15 equal regions on the dial
	
	
	for(sample = 0; sample < sampleFrames;sample++)
	{
		linFreq = powf(10.0f, (dspParameter[kFreq] * 3.0f) + 1.30103f)/getSampleRate();
		reson = powf(10.0f, (dspParameter[kReson] * 2.0f - 1.0f));
		switch(type)
		{
			case kType1PoleLP:
				c = exp(-2.0 * Pi * linFreq);
				outL[sample] = opY1L = inL[sample] + (opY1L - inL[sample]) * c;
				outR[sample] = opY1R = inR[sample] + (opY1R - inR[sample]) * c;
				break;
			case kType1PoleHP:
				c = exp(-2.0 * Pi * linFreq);
				opY1L = inL[sample] + (opY1L - inL[sample]) * c;
				opY1R = inR[sample] + (opY1R - inR[sample]) * c;
				outL[sample] = inL[sample] - opY1L;
				outR[sample] = inR[sample] - opY1R;
				break;
			case kType1stOrdAP:
				if(oldLinFreq != linFreq || type != oldType)
				{
					oldLinFreq = linFreq;
					oldType = type;
					double tf = tan(Pi * linFreq);
					c = (tf -1.0f)/(tf + 1.0f);
				}
				foY1L = c * (inL[sample] - foY1L) + foX1L;
				foX1L = inL[sample];
				outL[sample] = foY1L; 
				foY1R = c * (inR[sample] - foY1R) + foX1R;
				foX1R = inR[sample];
				outR[sample] = foY1R;
				break;
			case kType1stOrdLP:
				if(oldLinFreq != linFreq|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldType = type;
					double tf = tan(Pi * linFreq);
					c = (tf -1.0f)/(tf + 1.0f);
				}
				foY1L = c * (inL[sample] - foY1L) + foX1L;
				foX1L = inL[sample];
				outL[sample] = foY1L + foX1L; 
				foY1R = c * (inR[sample] - foY1R) + foX1R;
				foX1R = inR[sample];
				outR[sample] = foY1R + foX1R;
				break;
			case kType1stOrdHP:
				if(oldLinFreq != linFreq|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldType = type;
					double tf = tan(Pi * linFreq);
					c = (tf -1.0f)/(tf + 1.0f);
				}
				foY1L = c * (inL[sample] - foY1L) + foX1L;
				foX1L = inL[sample];
				outL[sample] = foX1L - foY1L; 
				foY1R = c * (inR[sample] - foY1R) + foX1R;
				foX1R = inR[sample];
				outR[sample] = foX1R - foY1R;
				break;
			case kType2ndOrdAP:
				if(oldLinFreq != linFreq || oldReson != reson|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					double bw = linFreq * reson; // when reson is 1, bw = freq - reson 10, bw = freq * 10
					double tf = tan(Pi * bw);
					c = (tf -1.0f)/(tf + 1.0f);
					d = -1.0 * cos(twoPi * linFreq);
				}
				soY0L = -1.0 * c * inL[sample] 
				+ d * (1.0 - c) * soX1L 
				+ soX2L
				- d * (1.0 - c) * soY1L 
				+ c * soY2L;
				soX2L = soX1L;
				soX1L = inL[sample];
				soY2L = soY1L;
				soY1L = soY0L;
				outL[sample] = soY0L;
				
				soY0R = -1.0 * c * inR[sample] 
				+ d * (1.0 - c) * soX1R 
				+ soX2R
				- d * (1.0 - c) * soY1R 
				+ c * soY2R;
				soX2R = soX1R;
				soX1R = inR[sample];
				soY2R = soY1R;
				soY1R = soY0R;
				outR[sample] = soY0R;				
				break;
			case kType2ndOrdBR:
				if(oldLinFreq != linFreq || oldReson != reson|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					double bw = linFreq * reson; // when reson is 1, bw = freq - reson 10, bw = freq/10
					double tf = tan(Pi * bw);
					c = (tf -1.0f)/(tf + 1.0f);
					d = -1.0 * cos(twoPi * linFreq);
				}
				soY0L = -1.0 * c * inL[sample] 
				+ d * (1.0 - c) * soX1L 
				+ soX2L
				- d * (1.0 - c) * soY1L 
				+ c * soY2L;
				soX2L = soX1L;
				soX1L = inL[sample];
				soY2L = soY1L;
				soY1L = soY0L;
				outL[sample] = soX1L + soY0L;
				
				soY0R = -1.0 * c * inR[sample] 
				+ d * (1.0 - c) * soX1R 
				+ soX2R
				- d * (1.0 - c) * soY1R 
				+ c * soY2R;
				soX2R = soX1R;
				soX1R = inR[sample];
				soY2R = soY1R;
				soY1R = soY0R;
				outR[sample] = soX1R + soY0R;
				break;
			case kType2ndOrdBP:
				if(oldLinFreq != linFreq || oldReson != reson|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					double bw = linFreq * reson; // when reson is 1, bw = freq - reson 10, bw = freq * 10
					double tf = tan(Pi * bw);
					c = (tf -1.0f)/(tf + 1.0f);
					d = -1.0 * cos(twoPi * linFreq);
				}
				soY0L = -1.0 * c * inL[sample] 
				+ d * (1.0 - c) * soX1L 
				+ soX2L
				- d * (1.0 - c) * soY1L 
				+ c * soY2L;
				soX2L = soX1L;
				soX1L = inL[sample];
				soY2L = soY1L;
				soY1L = soY0L;
				outL[sample] = soX1L - soY0L;
				
				soY0R = -1.0 * c * inR[sample] 
				+ d * (1.0 - c) * soX1R 
				+ soX2R
				- d * (1.0 - c) * soY1R 
				+ c * soY2R;
				soX2R = soX1R;
				soX1R = inR[sample];
				soY2R = soY1R;
				soY1R = soY0R;
				outR[sample] = soX1R - soY0R;
				break;
			case kTypeBQBWLP:
				if(oldLinFreq != linFreq || oldReson != reson || type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					if(reson > sqrt(2.0))
						reson = sqrt(2.0);
					if(reson < 0.1)
						reson = 0.1;
					if(linFreq < 0.1) 
						c = 1.0 / (linFreq * Pi);
					else 
						c = tan((0.5 - linFreq) * Pi);
				
					bqA1 = 1.0 / ( 1.0 + reson * c + c * c);
					bqA2 = 2.0 * bqA1;
					bqA3 = bqA1;
					bqB1 = 2.0 * ( 1.0 - c * c) * bqA1;
					bqB2 = ( 1.0 - reson * c + c * c) * bqA1;
				}
				bqY0L = bqA1 * inL[sample] + bqA2 * bqX1L + bqA3 * bqX2L - bqB1 * bqY1L - bqB2 * bqY2L;
				
				bqX2L = bqX1L;
				bqX1L = inL[sample];
				bqY2L = bqY1L;
				bqY1L = bqY0L;
				outL[sample] = bqY0L;
				
				bqY0R = bqA1 * inR[sample] + bqA2 * bqX1R + bqA3 * bqX2R - bqB1 * bqY1R - bqB2 * bqY2R;
				
				bqX2R = bqX1R;
				bqX1R = inR[sample];
				bqY2R = bqY1R;
				bqY1R = bqY0R;
				outR[sample] = bqY0R;
				
				break;
			case kTypeBQBWHP:
				if(oldLinFreq != linFreq || oldReson != reson || type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					if(reson > sqrt(2.0))
						reson = sqrt(2.0);
					if(reson < 0.1)
						reson = 0.1;
					c = tan(linFreq * Pi);
					
					bqA1 = 1.0 / ( 1.0 + reson * c + c * c);
					bqA2 = -2.0 * bqA1;
					bqA3 = bqA1;
					bqB1 = 2.0 * (c * c - 1.0) * bqA1;
					bqB2 = ( 1.0 - reson * c + c * c) * bqA1;
				}
				bqY0L = bqA1 * inL[sample] + bqA2 * bqX1L + bqA3 * bqX2L - bqB1 * bqY1L - bqB2 * bqY2L;
				
				bqX2L = bqX1L;
				bqX1L = inL[sample];
				bqY2L = bqY1L;
				bqY1L = bqY0L;
				outL[sample] = bqY0L;
				
				bqY0R = bqA1 * inR[sample] + bqA2 * bqX1R + bqA3 * bqX2R - bqB1 * bqY1R - bqB2 * bqY2R;
				
				bqX2R = bqX1R;
				bqX1R = inR[sample];
				bqY2R = bqY1R;
				bqY1R = bqY0R;
				outR[sample] = bqY0R;
				break;
			case kTypeSVLP:
				if(oldLinFreq != linFreq || oldReson != reson || type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					d = 2.0*sin(Pi * linFreq);
					if(d > 0.5)
						d = 0.5;
					c = reson;
					if(c > 2.0)
						c = 2.0;
					if(c > (2.0/d - (d * 0.5)))
						c = 2.0/d - (d * 0.5);
				}
				svBRL = inL[sample] - (c * svBPL);
				svLPL = svLPL + (d * svBPL);
				svHPL  = svBRL - svLPL;
				svBPL = svBPL + (d * svHPL);
				outL[sample] = svLPL;
				svBRR = inR[sample] - (c * svBPR);
				svLPR = svLPR + (d * svBPR);
				svHPR  = svBRR - svLPR;
				svBPR = svBPR + (d * svHPR);
				outR[sample] = svLPR;
				break;
			case kTypeSVHP:
				if(oldLinFreq != linFreq || oldReson != reson|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					d = 2.0*sin(Pi * linFreq);
					if(d > 0.5)
						d = 0.5;
					c = reson;
					if(c > 2.0)
						c = 2.0;
					if(c > (2.0/d - (d * 0.5)))
						c = 2.0/d - (d * 0.5);
				}
				svBRL = inL[sample] - (c * svBPL);
				svLPL = svLPL + (d * svBPL);
				svHPL  = svBRL - svLPL;
				svBPL = svBPL + (d * svHPL);
				outL[sample] = svHPL;
				svBRR = inR[sample] - (c * svBPR);
				svLPR = svLPR + (d * svBPR);
				svHPR  = svBRR - svLPR;
				svBPR = svBPR + (d * svHPR);
				outR[sample] = svHPR;
				break;
			case kTypeSVBR:
				if(oldLinFreq != linFreq || oldReson != reson|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					d = 2.0*sin(Pi * linFreq);
					if(d > 0.5)
						d = 0.5;
					c = reson;
					if(c > 2.0)
						c = 2.0;
					if(c > (2.0/d - (d * 0.5)))
						c = 2.0/d - (d * 0.5);
				}
				svBRL = inL[sample] - (c * svBPL);
				svLPL = svLPL + (d * svBPL);
				svHPL  = svBRL - svLPL;
				svBPL = svBPL + (d * svHPL);
				outL[sample] = svBRL;
				svBRR = inR[sample] - (c * svBPR);
				svLPR = svLPR + (d * svBPR);
				svHPR  = svBRR - svLPR;
				svBPR = svBPR + (d * svHPR);
				outR[sample] = svBRR;
				break;
			case kTypeSVBP:
				if(oldLinFreq != linFreq || oldReson != reson|| type != oldType)
				{
					oldLinFreq = linFreq;
					oldReson = reson;
					oldType = type;
					d = 2.0*sin(Pi * linFreq);
					if(d > 0.5)
						d = 0.5;
					c = reson;
					if(c > 2.0)
						c = 2.0;
					if(c > (2.0/d - (d * 0.5)))
						c = 2.0/d - (d * 0.5);
				}
				svBRL = inL[sample] - (c * svBPL);
				svLPL = svLPL + (d * svBPL);
				svHPL  = svBRL - svLPL;
				svBPL = svBPL + (d * svHPL);
				outL[sample] = svBPL;
				svBRR = inR[sample] - (c * svBPR);
				svLPR = svLPR + (d * svBPR);
				svHPR  = svBRR - svLPR;
				svBPR = svBPR + (d * svHPR);
				outR[sample] = svBPR;
				break;
			case kTypeBypass:
				outL[sample] = inL[sample];
				outR[sample] = inR[sample];
				break;
		}
		dspParameter[kFreq] += freqInc;
		dspParameter[kReson] += resonInc;
	}
	dspParameter[kFreq] = guiParameter[kFreq];
	dspParameter[kReson] = guiParameter[kReson];

}
